
/**
 * This class manages connections made to the server.
 *
 *
 * @version    $Id: ServerHelper.java,v 1.1 2000/08/14 23:56:47 jst3290 Exp $
 *
 * @author     Joe Testa
 *
 *
 * Revisions:
 *             $Log: ServerHelper.java,v $
 *             Revision 1.1  2000/08/14 23:56:47  jst3290
 *             Initial revision
 *
 */

package WebServer;
import java.util.*;
import java.net.*;
import java.io.*;
import Caching.*;
import HTTPMethods.*;
import HTTPClient.*;

import StartAllServers.Debug;
public class ServerHelper extends Thread {


    public final int READ_TIMEOUT = 15000;  // Max milliseconds for 'read()' to
                                            //     block.
    public final int READ_LIMIT = 10000;    // Max number of characters
                                            //     accepted from client.

    public static PuzzleBox puzzleBox;      // A reference to the preferences.


    private Socket connection;      // A reference to the current socket.
    private DataInputStream in;     // A stream to read from the socket.
    private DataOutputStream out;   // A stream to write to the socket.



    private File webRoot;           // The webroot, as set in the preferences
                                    //     file.
    private String firstLine;       // The first line of input from the client.
    private File requestedObject;   // The virtual object requested by the
                                    //     client.
    private String hostAddress;     // The IP address of the client.
    private boolean continueFlag;   // Flag that indicates if execution should
                                    //     continue.



    /**
     * Main constructor
     *
     * @param     connection     the socket to service.
     */

    public ServerHelper( Socket connection ) {

        this.connection = connection;

    }



    /**
     * The run method for this thread.
     */

    public void run() {


        // These operations are conducted here instead of the constructor so as
        // to not hold up the calling 'SimpleServer' object.

        String rawInput, physicalPath, virtualPath, methodType;

        continueFlag = true;
        webRoot = new File( puzzleBox.getWebRoot() );
        hostAddress = (connection.getInetAddress()).getHostAddress();


        // Check if the client is allowed to make a request.
        if ( !checkIPAddress() ) {
            killConnection();
            writeConsoleString( "Rejected IP Address." );
            return;
        }


        try {

            connection.setSoTimeout( READ_TIMEOUT );
            in = new DataInputStream( connection.getInputStream() );
            out = new DataOutputStream( connection.getOutputStream() );

        } catch( SocketException e ) {
            continueFlag = false;
            System.err.println( "Socket error:  " + e.getMessage() );
        } catch( IOException e ) {
            continueFlag = false;
            System.err.println( "I/O error:  " + e.getMessage() );
        }




        if ( !continueFlag )  return;


        // Read input from the socket.
        rawInput = readInput();


        if ( !continueFlag )  return;


        // Extract the first line of input.
        firstLine = getFirstLine( rawInput );
        SimpleServer.logAccess( hostAddress, new Date().toString(), firstLine );

        writeConsoleString( "Logged access: " + firstLine );


        if ( !continueFlag )  return;


        // Extract method type and virtual path from first line.
        methodType = getMethodType( firstLine );
        virtualPath = getVirtualPath( firstLine );


        // Only the GET method is supported.
        if ( !(methodType.toUpperCase()).equals( "GET" ) )
            errorOut( 501 );


        if ( !continueFlag )  return;



        // Determine where the requested object lies on the local
        //     file structure.  Send back a 403 error if outside webroot.

        requestedObject = mapObject( virtualPath );
        if ( !isSafe( requestedObject ) )  errorOut( 403 );


        if ( !continueFlag )  return;



        // If object does not exist, send back a 404 error.  If object cannot
        //     be read, send back a 403 error.

	System.out.println("The requested object was: " + requestedObject.toString());
        if ( !requestedObject.exists() )   {
	    System.out.println("VIRUAL PATH: " + virtualPath);
	    String fileName = requestedObject.toString();
	    String urlname = null;

	    //remove unwanted URL escape characters
	    try{
		urlname = Codecs.URLDecode(fileName);
		virtualPath = Codecs.URLDecode(virtualPath);
		System.out.println("VIRTUAL PATH: " + virtualPath);
		System.out.println("URLNAME: " + urlname);
	    }catch(Exception e){
		System.out.println(e);
	    }

	    /*******************ADDED CODE*****************/
	    //Convert request to MD5 format and check if it exists
	    urlname=ParseRequest.attachIndexHtml(urlname);

	    int index = urlname.indexOf('/');
	    int index2 = urlname.lastIndexOf('/');
	    String uName = urlname.substring(index+1, index2);
	    String uName2 = urlname.substring(index2+1, urlname.length());

	    MD5 md5 = new MD5();

	    String dirName = (".cache/").concat(md5.toDigest(uName));
	    fileName = dirName + "/" +uName2;
	    /*******************************************/

	    File file = new File(fileName);
	    if(!file.exists())
		errorOut( 404 );
	    else
		requestedObject = file;
	}
        if ( !continueFlag )  return;
	
        if ( !requestedObject.canRead() )  errorOut( 403 );
        if ( !continueFlag )  return;
	


        // If object is a file, send it to the user, if object is directory,
        // call 'processDirectory' (which does not necessarily list its files
        // automatically).

        if ( requestedObject.isFile() )
            sendFile( requestedObject, virtualPath );
        else if ( requestedObject.isDirectory() )
            processDirectory( requestedObject, virtualPath );


    }



    /**
     * Checks IP address and determines if it should be allowed to make
     *     request.
     *
     * @return                  'true' if IP is allowed; 'false' if not.
     */

    private boolean checkIPAddress() {


        // Set this to true in case ipallow and ipdeny are undefined.
        boolean retVal = true;

        LinkedList theList;
        String allowedIP, disallowedIP, aString;
        StringTokenizer ipToken, ipToken2;
        String[] remoteIPFields = new String[ 4 ];
        int fieldMatches;


        // Break up the client's IP address by its class fields.
        ipToken = new StringTokenizer( hostAddress, "." );
        for ( int i = 0; i < 4; i++ )
            remoteIPFields[ i ] = ipToken.nextToken();



        // Set the list to the list of allowed IPs and check if there are any
        //     elements.

        theList = ( LinkedList )puzzleBox.getIPAllow();
        if ( theList.size() != 0 ) {


            // Guilty until proven innocent.
            retVal = false; 
            for ( int i = 0; ( i < theList.size() ) &&
                             ( retVal == false ); i++ ) {


                allowedIP = ( String )theList.get( i );
                ipToken2 = new StringTokenizer( allowedIP, "." );
                fieldMatches = 0;


                // Sift through all the IP fields and compare to client IP.
                //     Stars are accepted as wildcards.

                for ( int ii = 0; ii < 4; ii++ ) {
                    aString = ipToken2.nextToken();
                    if ( aString.equals( "*" ) )  fieldMatches++;
                    else if ( aString.equals( remoteIPFields[ ii ] ) )
                        fieldMatches++;
                }


                if ( fieldMatches == 4 )  retVal = true;

            }


        // There are no IPs specified in ipallow, so check ipdeny...
        } else {


            // Innocent until proven guilty.
            retVal = true;
            theList = ( LinkedList )puzzleBox.getIPDeny();
            if ( theList.size() != 0 ) {
                for ( int i = 0;
                      ( i < theList.size() ) && ( retVal == true ); i++ ) {


                    disallowedIP = ( String )theList.get( i );
                    ipToken2 = new StringTokenizer( disallowedIP, "." );
                    fieldMatches = 0;


                    // Sift through all the IP fields and compare to client IP.
                    //     Stars are accepted as wildcards.

                    for ( int ii = 0; ii < 4; ii++ ) {
                        aString = ipToken2.nextToken();
                        if ( aString.equals( "*" ) )  fieldMatches++;
                        else if ( aString.equals( remoteIPFields[ ii ] ) )
                            fieldMatches++;
                    }

                    if ( fieldMatches == 4 ) retVal = false;
                }
            }

        }

        return retVal;

    }



    /**
     * Sends an error to the error log, then to the socket, then the connection
     *     is killed.
     *
     * @param     errorCode     the error code.
     */

    private void errorOut( int errorCode ) {


        // Determine the error text based on the error code.
        String errorText = "";
        if ( errorCode == 403 )       errorText = "Forbidden";
        else if ( errorCode == 404 )  errorText = "Not Found";
        else if ( errorCode == 500 )  errorText = "Internal Server Error";
        else if ( errorCode == 501 )  errorText = "Not Implemented";


        // Log the error.
        SimpleServer.logError( hostAddress, new Date().toString(), errorCode,
                                                        errorText, firstLine );
        writeConsoleString( "Logged error: " + errorCode );





        writeSocketString( "HTTP/0.9 " + errorCode + " " + errorText +
                                                                  "\r\n\r\n" );
        writeSocketString( parseErrorTemplate( errorCode, errorText ) );
        killConnection();


    }



    /**
     * Extracts the first line from input.
     *
     * @param      rawInput      input from which the first line is taken.
     * @return                   the first line.
     */

    private String getFirstLine( String rawInput ) {


        StringTokenizer wholeInput = new StringTokenizer( rawInput, "\r\n" );
        String retVal;

        if ( !wholeInput.hasMoreTokens() ) {
            retVal = "";
            continueFlag = false;
        } else
            retVal = wholeInput.nextToken();

        return retVal;


    }



    /**
     * Extracts the method from the first line of input.
     *
     * @param     firstLine     the first line of input from the socket.
     * @return                      the method type.
     */

    private String getMethodType( String firstLine ) {


        StringTokenizer methodAndPath;
        String retVal, method;
     
        methodAndPath = new StringTokenizer( firstLine );
        if ( !methodAndPath.hasMoreTokens() ) {
            retVal = "";
            continueFlag = false;
        } else
            retVal = methodAndPath.nextToken();

        
        return retVal;


    }



    /**
     * Returns the mime type of the specified file (using the PuzzleBox
     *     object).
     *
     * @param     physicalPath      the file whose mime type is to be
     *                                  determined.
     * @return                      the file's mime type (empty if
     *                                  unsupported).
     */

    private String getMIMEType( File requestedObject ) {


        String retVal = "";
        String filePath = "", suffix = "";
        int periodPosition = -1;


        try {


            filePath = requestedObject.getCanonicalPath();

            periodPosition = filePath.lastIndexOf( "." );
            if ( periodPosition != -1 ) {
                suffix = filePath.substring( periodPosition + 1 );
                retVal = puzzleBox.getMIMEType( suffix.toLowerCase() );
                if ( retVal == null )  retVal = "unknown";
                else if ( retVal.equals( "" ) )  retVal = "unknown";
            } else
                retVal = "text/plain";


        } catch( IOException e )  {  retVal = ""; continueFlag = false;  }


        return retVal;


    }



    /**
     * Extracts the virtual path from the first line of input.
     *
     * @param      firstLine     the first line of input from the socket.
     * @return                   the virtual path.
     */

    private String getVirtualPath( String firstLine ) {


        StringTokenizer methodAndPath;
        String method, retVal = "";

        methodAndPath = new StringTokenizer( firstLine );

        if ( methodAndPath.hasMoreTokens() ) 
            method = methodAndPath.nextToken();
      


        if ( methodAndPath.hasMoreTokens() )
            retVal = methodAndPath.nextToken();
        else
            continueFlag = false;


        return retVal;


    }



    /**
     * Closes the socket.
     */

    public void killConnection() {

        continueFlag = false;

        try {  
            if ( connection != null )  connection.close();
        } catch( IOException e ) {}


    }



    /**
     * My Kung-Fu is the best!  My Dragon technique defeat your Tiger
     *     technique!
     *
     * @param    requestedObject    the file to check.
     * @return                      'true' if the object is within legal
     *                                  bounds; 'false' if otherwise.
     */

    private boolean isSafe( File requestedObject ) {


        boolean retVal = false;

        try {
            if ( (requestedObject.getCanonicalPath()).startsWith( 
                                                 webRoot.getCanonicalPath() ) )
                retVal = true;
        } catch( IOException e ) {  retVal = false; continueFlag = false;  }

        return retVal;


    }



    /**
     * This method creates the necessary header information.  Pretty useless
     *     under HTTP/0.9, but will be expanded for HTTP/1.0+.
     *
     * @param    fileLength    the number to set the 'Content-Length' field.
     * @param    mimeType      the mime type of the content.
     * @return                 the header.
     */

    private String makeHeader( int fileLength, String mimeType ) {


        String header = "";
        header += ( "HTTP/1.1 200 OK\r\n" );
        header += ( "Content-Length: " + fileLength + "\r\n" );
        header += ( "Content-Type: " + mimeType + "\r\n" );

        //header += ( "Date: " + Util.httpDate( new Date() ) + "\r\n" );
	header += ( "Date: " + Util.httpDate(Calendar.getInstance().getTime()) + "\r\n" );

	CacheBuffer cb = CacheBuffer.getInstance();
	String virtualPath = getVirtualPath( firstLine );

	//I have hardcoded the path
	//System.out.println("*******The virtual path is: " + virtualPath);
	if (virtualPath.startsWith("/")) {
	    virtualPath = "http://" + virtualPath.substring(1, virtualPath.length());
	    //System.out.println("The corrected VirtualPath is: " + virtualPath);
	} else 
	    virtualPath = "http://" + virtualPath;

	BufferEntry entry = cb.getBufferEntry(virtualPath);

	if (entry != null) {
	    Date last_modified = entry.getModified();
	    Date expires = entry.getExpired();
	
	    if (last_modified != null)
		header += ( "Last-Modified: " + Util.httpDate(last_modified) + "\r\n" );
	    if (expires != null) 
		header += ( "Expires: " +Util.httpDate(expires) + "\r\n" );
	    

	}
	header += ( "Server: " + SimpleServer.PROGRAM_NAME + "\r\n\r\n" );
        return header;


    }



    /**
     * Maps a virtual object to a physical path.
     *
     * @param     virtualPath      a virtualPath whose physical location is
     *                                 to be determined.
     * @return                     the physical path of the object.
     */

    private File mapObject( String virtualPath ) {
        return( new File( webRoot, virtualPath ) );
    }



    /**
     * Inserts variables into error template.
     *
     * @param    errorCode     the error code in decimal format.
     * @param    errorText   the error code in string format.
     */

    private String parseErrorTemplate( int errorCode, String errorText ) {

        
        BufferedReader in = null;
        String totalContents = "", aLine = "";
        int stringPos = 0;

        try {

            // Open the error template, read it line by line and save it
            //     in 'totalContents'.
            in = new BufferedReader( new FileReader(
                                              puzzleBox.getErrorTemplate() ) );
            while ( ( aLine = in.readLine() ) != null )
                totalContents += ( aLine + "\r\n" );
            in.close();

        } catch ( FileNotFoundException e ) {
            System.err.println( e.getMessage() );
        } catch ( IOException e ) {
	    if (Debug.DEBUG_WebServer == true) { 
		System.out.println( "I/O error:  " + e.getMessage() ); 
	    }
        }


        // Replace the error template variables with their respective data.
        totalContents = replace( totalContents, "%error_code",
                                     String.valueOf( errorCode ) );
        totalContents = replace( totalContents, "%error_text", errorText );
        totalContents = replace( totalContents, "%server_version",
                                     SimpleServer.PROGRAM_NAME );
        totalContents = replace( totalContents, "%webmaster_email",
                                     puzzleBox.getWebmasterEmail() );
        totalContents = replace( totalContents, "%hostname",
                                     puzzleBox.getHostName() );
        totalContents = replace( totalContents, "%port",
                                     String.valueOf( puzzleBox.getPort() ) );
        totalContents = replace( totalContents, "%remote_ip",
                                     hostAddress );
        totalContents = replace( totalContents, "%random_picture",
                                     SimpleServer.getRandomPicture() );
       
        return totalContents;


    }



    /**
     * This method checks for file named "index.html" in the directory, and 
     * returns it if it exists.  Otherwise, a list of files in the directory
     * is returned if directory browsing is enabled.
     * 
     * @param     physicalPath    the directory to process.
     * @param     virtualPath     the virtual path.  Used only in error message
     *                                if directory browsing is disabled.
     */

    private void processDirectory( File requestedObject, String virtualPath ) {


        // If a file named "index.html" exists and is readable,
        //     send that file and close the connection.

        File indexFile = new File( requestedObject, "index.html" );
        if ( indexFile.exists() && indexFile.canRead() )
            sendFile( indexFile, virtualPath );
        else if ( puzzleBox.getDirectoryBrowsing() ) {  // Check if a directory
                                                        //     listing should
                                                        //     be sent.


            String[] files;
            String directoryListing, fileURLPrefix, imageURL, hostName;
            int port;

            if ( virtualPath.charAt( virtualPath.length() - 1 ) != '/' )
                virtualPath += "/";


            port = puzzleBox.getPort();
            hostName = puzzleBox.getHostName();
            if ( port != 80 ) {
                fileURLPrefix = "http://" + hostName + ":" + port +
                                                   virtualPath;
                imageURL = "http://" + hostName + ":" + port + "/images/" +
                                                   SimpleServer.getRandomPicture();
            } else {
                fileURLPrefix = "http://" + hostName + virtualPath;
                imageURL = "http://" + hostName + "/images/" +
                                                   SimpleServer.getRandomPicture();
            }


            directoryListing  = "<html><head><title>Directory Listing Of " +
                                                  virtualPath + "</title>\r\n";
            directoryListing += "<style type=\"text/css\">\r\n";
            directoryListing += "\tBODY { background: black; }\r\n";
            directoryListing += "\tA.files { color: rgb( 255, 79, 79 ); }\r\n";
            directoryListing += "\tA.program { color:  red; }\r\n";
            directoryListing +=
                           "\tP { color: red;  font-family: \"symbol\"; }\r\n";
            directoryListing += "</style></head>\r\n\r\n";


            directoryListing += "<body><br><br><br><center><p>Contents of " +
                                    virtualPath + ":</p></center><br><br>\r\n";
            directoryListing += "<pre>\r\n";


            files = requestedObject.list();
            for ( int i = 0; i < files.length; i++ )
                directoryListing += ( "        <a href=\"" +
                                      fileURLPrefix + files[ i ] +
                                      "\" class=\"files\">" +  files[ i ] +
                                      "</a>\r\n" );

            directoryListing += ( "</pre>\r\n" );
            directoryListing += ( "<br><br><center><a href=\"" +
                                "http://www.rit.edu/~jst3290/\" class=\"" +
                                "program\"><img src=\"" + imageURL +
                                "\" alt=\"" + SimpleServer.PROGRAM_NAME +
                                "\" border=0></a></center>\r\n\r\n" );
            directoryListing += "</html>";

            //writeSocketString( makeHeader( directoryListing.length(),
	    //"text/html" ) + directoryListing );
            killConnection();


        } else                    // An index.html' file doesn't exist in the
            errorOut( 403 );      //     directory, and listing is not enabled,
                                  //      so send a "403 Forbidden" error.

    }



    /**
     * Read input from the socket.  This method will return an empty string if
     *     the timeout was reached, or an I/O error was raised.
     *
     * @return            raw input from the socket.
     */

    private String readInput() {


        byte[] byteArray;
        String doubleEnter = "\r\n\r\n";
        String retVal = "";
        int available;
        long startTime;


        // Read from the socket.  If READ_TIMEOUT milliseconds elapse, or an IO
        //     Exception is thrown, return an empty string, and signal for
        //     this thread to die.
        try {

            startTime = System.currentTimeMillis();
            while( continueFlag &&
                  ( System.currentTimeMillis() - startTime < READ_TIMEOUT ) ) {


                // Check to see how many bytes are available in this socket.
                //     If there is nothing to read, then skip down to the end
                //     of the 'while()' loop.
                available = in.available();
                if ( available < 1 )  continue;


                // If the client has not overflowed the read limit, then keep
                //     reading from the socket.
                if ( available + retVal.length() < READ_LIMIT ) {

                    byteArray = new byte[ available ];
                    in.read( byteArray );
                    retVal += new String( byteArray );

                } else {


                    // The client is trying to flood the server, so kill the
                    //     connection and write to the error log.  Note, this
                    //     part is executed if the the client *attempted* to
                    //     send more characters than allowed -- it does not
                    //     necessarily mean that they were all read into the
                    //     buffer from the socket.

                    killConnection();
                    String theDate = ( new Date() ).toString();
                    SimpleServer.logAccess( hostAddress, theDate,
                                                         "[ see error log ]" );
                    SimpleServer.logError( hostAddress, theDate, 500,
                             "Input greater than " + READ_LIMIT + " received.",
                             retVal.substring( 0, 100 ) );
                    writeConsoleString( "Logged error: 500; client exceeded " +
                                        "read limit of " + READ_LIMIT + "." );
                    continue;

                }


                // If the last four characters sent by the client are CR LF,
                //     CR LF, then it is done sending input to the server.
                if ( retVal.length() >= 4 )
                    if ( retVal.substring( retVal.length() - 4 ).equals( 
                                                        doubleEnter ) )  break;


            }

        } 
        catch( InterruptedIOException e ) {  retVal = "";  }
        catch( IOException e )            {  retVal = "";  }


        // Kill the connection if it was idle, if READ_TIMEOUT elapsed, or if
        //     an I/O error occcured.
        if ( retVal.equals( "" ) )  killConnection();

        return retVal;


    }



    /**
     * Replaces all instances of one string with another.  Used to expand
     *     error template variables with values.
     *
     * @param    theString     the string to search through.
     * @param    oldText       the string to replace.
     * @param    newText       the string to replace 'oldText' with.
     * @return                 the translated string.
     */

    private String replace( String theString,
                            String oldText,
                            String newText ) {


        String retVal = theString.substring( 0 );

        String part1 = "", part2 = "";
        int stringPos = -1;


        while( ( stringPos = retVal.indexOf( oldText ) ) != -1 )  {
            part1 = retVal.substring( 0, stringPos );
            part2 = retVal.substring( stringPos + oldText.length() );
            retVal = part1 + newText + part2;
        }

        return retVal;


    }



    /**
     * Sends a file through the socket.
     *
     * @param      requestedObject      the file to send.
     * @param      virtualPath          the virtual object.  Used in error
     *                                      message when an exeception is
     *                                      thrown while processing file (to
     *                                      keep server's internal structure
     *                                      hidden).
     */

    private void sendFile( File requestedObject, String virtualPath ) {


        String mimeType = getMIMEType( requestedObject );
	System.out.println("SEND FILE " + mimeType);

        // The PuzzleBox class may be a little buggy...
        if ( mimeType.equals( "" ) )
            errorOut( 500 );
	/*        else if ( mimeType.equals( "unknown" ) )
            errorOut( 501 );
	*/



        try {

            FileInputStream fIn;
            byte[] fContents;
            int fileLength;


            // Get the size of the file.
            fileLength = ( int )requestedObject.length();
	    System.out.println("FILE LENGTH: " + fileLength);

            // Read the file using a byte array and send it through the socket.
            fIn = new FileInputStream( requestedObject );
            fContents = new byte[ fileLength + 1 ];
            fIn.read( fContents );
            fIn.close();

            writeSocketString( makeHeader( fileLength, mimeType ) );
	    //writeSocketString("\r\n\r\n\r\n\r\n");
            writeSocketBytes( fContents );

        } catch( IOException e ) {}


        killConnection();


    }



    /**
     * Writes a string to the console if the verbose flag is set.
     *
     * @param     output    the string to write.
     */

    private void writeConsoleString( String output ) {

        if ( SimpleServer.verboseFlag )
	    System.out.println( "    - [" + hostAddress + "]  " + output ); 
 	

    }



    /**
     * Writes an array of bytes through the socket.
     *
     * @param     output    the bytes to write.
     */

    private void writeSocketBytes( byte[] output ) {

        try {
            if ( out != null )  out.write( output );
            else                continueFlag = false;
        }
        catch( IOException e ) {  continueFlag = false;  }

    }



    /**
     * Writes a string through the socket.
     *
     * @param     output    the string to write.
     */

    private void writeSocketString( String output ) {

        try {
            if ( out != null )  out.writeBytes( output );
            else                continueFlag = false;
        }
        catch( IOException e ) {  continueFlag = false;  }

    }



}  // ServerHelper






